@@ -197,16 +197,21 @@ PythonStackTracker::emitPendingPushesAndPops()
     }
     auto first_to_emit = it.base();
 
-    // Emit pending pops
-    Tracker::getTracker()->popFrames(d_num_pending_pops);
-    d_num_pending_pops = 0;
+    Tracker* tracker = Tracker::getTracker();
+    if (tracker) {
+        // Emit pending pops
+        if (d_num_pending_pops) {
+            tracker->popFrames(d_num_pending_pops);
+            d_num_pending_pops = 0;
+        }
 
-    // Emit pending pushes
-    for (auto to_emit = first_to_emit; to_emit != d_stack->end(); ++to_emit) {
-        if (!Tracker::getTracker()->pushFrame(to_emit->raw_frame_record)) {
-            break;
+        // Emit pending pushes
+        for (auto to_emit = first_to_emit; to_emit != d_stack->end(); ++to_emit) {
+            if (!tracker->pushFrame(to_emit->raw_frame_record)) {
+                break;
+            }
+            to_emit->state = FrameState::EMITTED_AND_LINE_NUMBER_HAS_NOT_CHANGED;
         }
-        to_emit->state = FrameState::EMITTED_AND_LINE_NUMBER_HAS_NOT_CHANGED;
     }
 
     invalidateMostRecentFrameLineNumber();
@@ -420,11 +425,10 @@ PythonStackTracker::handleGreenletSwitch(PyObject* from, PyObject* to)
     std::for_each(stack.rbegin(), stack.rend(), [this](auto& frame) { pushPythonFrame(frame); });
 }
 
-std::atomic<bool> Tracker::d_active = false;
-std::unique_ptr<Tracker> Tracker::d_instance_owner;
-std::atomic<Tracker*> Tracker::d_instance = nullptr;
-
-MEMRAY_FAST_TLS thread_local size_t NativeTrace::MAX_SIZE{128};
+std::unique_ptr<std::mutex> Tracker::s_mutex(new std::mutex);
+pthread_key_t Tracker::s_native_unwind_vector_key;
+std::unique_ptr<Tracker> Tracker::s_instance_owner;
+std::atomic<Tracker*> Tracker::s_instance = nullptr;
 
 std::vector<PythonStackTracker::LazilyEmittedFrame>
 PythonStackTracker::pythonFrameToStack(PyFrameObject* current_frame)
@@ -545,11 +549,20 @@ Tracker::Tracker(
 , d_follow_fork(follow_fork)
 , d_trace_python_allocators(trace_python_allocators)
 {
-    // Note: this must be set before the hooks are installed.
-    d_instance = this;
-
     static std::once_flag once;
     call_once(once, [] {
+        // We use the pthread TLS API for this vector because we must be able
+        // to re-create it while TLS destructors are running (a destructor can
+        // call malloc, hitting our malloc hook). POSIX guarantees multiple
+        // rounds of TLS destruction if destructors call pthread_setspecific.
+        // Note: If this raises an exception, the call_once can be retried.
+        if (0 != pthread_key_create(&s_native_unwind_vector_key, [](void* data) {
+                delete static_cast<std::vector<NativeTrace::ip_t>*>(data);
+            }))
+        {
+            throw std::runtime_error{"Failed to create pthread key"};
+        }
+
         hooks::ensureAllHooksAreValid();
         NativeTrace::setup();
 
@@ -562,9 +575,10 @@ Tracker::Tracker(
     if (!d_writer->writeHeader(false)) {
         throw IoError{"Failed to write output header"};
     }
-    updateModuleCache();
 
     RecursionGuard guard;
+    updateModuleCacheImpl();
+
     PythonStackTracker::s_native_tracking_enabled = native_traces;
     PythonStackTracker::installProfileHooks();
     if (d_trace_python_allocators) {
@@ -574,34 +588,39 @@ Tracker::Tracker(
     d_background_thread->start();
 
     d_patcher.overwrite_symbols();
-
-    tracking_api::Tracker::activate();
 }
 
 Tracker::~Tracker()
 {
     RecursionGuard guard;
     tracking_api::Tracker::deactivate();
+
     PythonStackTracker::s_native_tracking_enabled = false;
     d_background_thread->stop();
-    d_patcher.restore_symbols();
+
+    {
+        std::scoped_lock<std::mutex> lock(*s_mutex);
+        d_patcher.restore_symbols();
+    }
+
     if (Py_IsInitialized() && !_Py_IsFinalizing()) {
         PyGILState_STATE gstate;
         gstate = PyGILState_Ensure();
 
         if (d_trace_python_allocators) {
+            std::scoped_lock<std::mutex> lock(*s_mutex);
             unregisterPymallocHooks();
         }
+
         PythonStackTracker::removeProfileHooks();
 
         PyGILState_Release(gstate);
     }
+
+    std::scoped_lock<std::mutex> lock(*s_mutex);
     d_writer->writeTrailer();
     d_writer->writeHeader(true);
     d_writer.reset();
-
-    // Note: this must not be unset before the hooks are uninstalled.
-    d_instance = nullptr;
 }
 
 Tracker::BackgroundThread::BackgroundThread(
@@ -673,15 +692,20 @@ Tracker::BackgroundThread::start()
                     break;
                 }
             }
+
             size_t rss = getRSS();
             if (rss == 0) {
                 Tracker::deactivate();
                 break;
             }
-            if (!d_writer->writeRecord(MemoryRecord{timeElapsed(), rss})) {
-                std::cerr << "Failed to write output, deactivating tracking" << std::endl;
-                Tracker::deactivate();
-                break;
+
+            {
+                std::lock_guard<std::mutex> lock(*s_mutex);
+                if (!d_writer->writeRecord(MemoryRecord{timeElapsed(), rss})) {
+                    std::cerr << "Failed to write output, deactivating tracking" << std::endl;
+                    Tracker::deactivate();
+                    break;
+                }
             }
         }
     });
@@ -693,8 +717,8 @@ Tracker::BackgroundThread::stop()
     {
         std::scoped_lock<std::mutex> lock(d_mutex);
         d_stop = true;
-        d_cv.notify_one();
     }
+    d_cv.notify_one();
     if (d_thread.joinable()) {
         try {
             d_thread.join();
@@ -726,37 +750,43 @@ Tracker::childFork()
     // exists, and potentially to flush buffered output to a socket it no
     // longer owns. Note that d_instance_owner is always set after d_instance
     // and unset before d_instance.
-    (void)d_instance_owner.release();
+    (void)s_instance_owner.release();
 
-    Tracker* old_tracker = d_instance;
+    // Likewise, leak our old mutex, and re-create it.
+    (void)s_mutex.release();
+    s_mutex.reset(new std::mutex);
+
+    // Save a reference to the old tracker (if any), then unset our singleton.
+    Tracker* old_tracker = s_instance;
+    Tracker::deactivate();
 
     // If we inherited an active tracker, try to clone its record writer.
     std::unique_ptr<RecordWriter> new_writer;
-    if (old_tracker && memray::tracking_api::Tracker::isActive() && old_tracker->d_follow_fork) {
+    if (old_tracker && old_tracker->d_follow_fork) {
         new_writer = old_tracker->d_writer->cloneInChildProcess();
     }
 
     if (!new_writer) {
-        // We either have no tracker, or a deactivated tracker, or a tracker
-        // with a sink that can't be cloned. Unset our singleton and bail out.
-        // Note that the old tracker's hooks may still be installed. This is
-        // OK, as long as they always check the (static) isActive() flag before
-        // calling any methods on the now null tracker singleton.
-        Tracker::deactivate();
-        d_instance = nullptr;
+        // Either tracking wasn't active, or the tracker was using a sink that
+        // can't be cloned. Leave our singleton unset and bail out. Note that
+        // the old tracker's hooks may still be installed.  If this process
+        // exits, trackDeallocation will be called to track the deallocation of
+        // s_mutex when the process's globals are destroyed! To handle this,
+        // the hooks must check the (static) isActive() flag before acquiring
+        // s_mutex or calling any methods on the now null tracker singleton.
         RecursionGuard::isActive = false;
         return;
     }
 
     // Re-enable tracking with a brand new tracker.
     // Disable tracking until the new tracker is fully installed.
-    Tracker::deactivate();
-    d_instance_owner.reset(new Tracker(
+    s_instance_owner.reset(new Tracker(
             std::move(new_writer),
             old_tracker->d_unwind_native_frames,
             old_tracker->d_memory_interval,
             old_tracker->d_follow_fork,
             old_tracker->d_trace_python_allocators));
+    Tracker::activate();
     RecursionGuard::isActive = false;
 }
 
@@ -780,24 +810,30 @@ Tracker::computeMainTidSkip()
     return num_frames - 1;
 }
 
-void
-Tracker::trackAllocationImpl(void* ptr, size_t size, hooks::Allocator func)
+bool
+Tracker::areNativeTracesEnabled()
 {
-    if (RecursionGuard::isActive || !Tracker::isActive()) {
-        return;
-    }
-    RecursionGuard guard;
+    return PythonStackTracker::s_native_tracking_enabled;
+}
 
+void
+Tracker::trackAllocationImpl(
+        void* ptr,
+        size_t size,
+        hooks::Allocator func,
+        const std::optional<NativeTrace>& trace)
+{
     PythonStackTracker::get().emitPendingPushesAndPops();
 
     if (d_unwind_native_frames) {
-        NativeTrace trace;
         frame_id_t native_index = 0;
+
         // Skip the internal frames so we don't need to filter them later.
-        if (trace.fill(2)) {
-            native_index = d_native_trace_tree.getTraceIndex(trace, [&](frame_id_t ip, uint32_t index) {
-                return d_writer->writeRecord(UnresolvedNativeFrame{ip, index});
-            });
+        if (trace && trace.value().size()) {
+            native_index =
+                    d_native_trace_tree.getTraceIndex(trace.value(), [&](frame_id_t ip, uint32_t index) {
+                        return d_writer->writeRecord(UnresolvedNativeFrame{ip, index});
+                    });
         }
         NativeAllocationRecord record{reinterpret_cast<uintptr_t>(ptr), size, func, native_index};
         if (!d_writer->writeThreadSpecificRecord(thread_id(), record)) {
@@ -816,11 +852,6 @@ Tracker::trackAllocationImpl(void* ptr, size_t size, hooks::Allocator func)
 void
 Tracker::trackDeallocationImpl(void* ptr, size_t size, hooks::Allocator func)
 {
-    if (RecursionGuard::isActive || !Tracker::isActive()) {
-        return;
-    }
-    RecursionGuard guard;
-
     AllocationRecord record{reinterpret_cast<uintptr_t>(ptr), size, func};
     if (!d_writer->writeThreadSpecificRecord(thread_id(), record)) {
         std::cerr << "Failed to write output, deactivating tracking" << std::endl;
@@ -831,16 +862,15 @@ Tracker::trackDeallocationImpl(void* ptr, size_t size, hooks::Allocator func)
 void
 Tracker::invalidate_module_cache_impl()
 {
-    RecursionGuard guard;
     d_patcher.overwrite_symbols();
-    updateModuleCache();
+    updateModuleCacheImpl();
 }
 
 #ifdef __linux__
 static int
 dl_iterate_phdr_callback(struct dl_phdr_info* info, [[maybe_unused]] size_t size, void* data)
 {
-    auto writer = reinterpret_cast<RecordWriter*>(data);
+    auto& mappings = *reinterpret_cast<std::vector<ImageSegments>*>(data);
     const char* filename = info->dlpi_name;
     std::string executable;
     assert(filename != nullptr);
@@ -861,20 +891,7 @@ dl_iterate_phdr_callback(struct dl_phdr_info* info, [[maybe_unused]] size_t size
         }
     }
 
-    if (!writer->writeRecordUnsafe(SegmentHeader{filename, segments.size(), info->dlpi_addr})) {
-        std::cerr << "memray: Failed to write output, deactivating tracking" << std::endl;
-        Tracker::deactivate();
-        return 1;
-    }
-
-    for (const auto& segment : segments) {
-        if (!writer->writeRecordUnsafe(segment)) {
-            std::cerr << "memray: Failed to write output, deactivating tracking" << std::endl;
-            Tracker::deactivate();
-            return 1;
-        }
-    }
-
+    mappings.push_back({filename, info->dlpi_addr, std::move(segments)});
     return 0;
 }
 #endif
@@ -885,14 +902,14 @@ Tracker::updateModuleCacheImpl()
     if (!d_unwind_native_frames) {
         return;
     }
-    auto writer_lock = d_writer->acquireLock();
-    if (!d_writer->writeRecordUnsafe(MemoryMapStart{})) {
-        std::cerr << "memray: Failed to write output, deactivating tracking" << std::endl;
-        deactivate();
-    }
+
+    static size_t s_last_mappings_size = 20;
+
+    std::vector<ImageSegments> mappings;
+    mappings.reserve(s_last_mappings_size + 1);
 
 #ifdef __linux__
-    dl_iterate_phdr(&dl_iterate_phdr_callback, d_writer.get());
+    dl_iterate_phdr(&dl_iterate_phdr_callback, &mappings);
 #elif defined(__APPLE__)
     uint32_t c = _dyld_image_count();
     for (uint32_t i = 0; i < c; i++) {
@@ -910,26 +927,23 @@ Tracker::updateModuleCacheImpl()
             }
         }
 
-        if (!d_writer->writeRecordUnsafe(SegmentHeader{image_name, segments.size(), slide})) {
-            std::cerr << "memray: Failed to write output, deactivating tracking" << std::endl;
-            Tracker::deactivate();
-            return;
-        }
-
-        for (const auto& segment : segments) {
-            if (!d_writer->writeRecordUnsafe(segment)) {
-                std::cerr << "memray: Failed to write output, deactivating tracking" << std::endl;
-                Tracker::deactivate();
-                return;
-            }
-        }
+        mappings.push_back({image_name, slide, std::move(segments)});
     }
 #endif
+
+    s_last_mappings_size = mappings.size();
+
+    if (!d_writer->writeMappings(mappings)) {
+        std::cerr << "memray: Failed to write output, deactivating tracking" << std::endl;
+        Tracker::deactivate();
+        return;
+    }
 }
 
 void
 Tracker::registerThreadNameImpl(const char* name)
 {
+    RecursionGuard guard;
     if (!d_writer->writeThreadSpecificRecord(thread_id(), ThreadRecord{name})) {
         std::cerr << "memray: Failed to write output, deactivating tracking" << std::endl;
         deactivate();
@@ -978,19 +992,20 @@ Tracker::pushFrame(const RawFrame& frame)
 void
 Tracker::activate()
 {
-    d_active = true;
+    // NOTE: Tracker::s_mutex must be held
+    s_instance = s_instance_owner.get();
 }
 
 void
 Tracker::deactivate()
 {
-    d_active = false;
+    s_instance = nullptr;
 }
 
-const std::atomic<bool>&
+bool
 Tracker::isActive()
 {
-    return Tracker::d_active;
+    return s_instance != nullptr;
 }
 
 // Static methods managing the singleton
@@ -1004,12 +1019,15 @@ Tracker::createTracker(
         bool trace_python_allocators)
 {
     // Note: the GIL is used for synchronization of the singleton
-    d_instance_owner.reset(new Tracker(
+    s_instance_owner.reset(new Tracker(
             std::move(record_writer),
             native_traces,
             memory_interval,
             follow_fork,
             trace_python_allocators));
+
+    std::unique_lock<std::mutex> lock(*s_mutex);
+    tracking_api::Tracker::activate();
     Py_RETURN_NONE;
 }
 
@@ -1017,14 +1035,14 @@ PyObject*
 Tracker::destroyTracker()
 {
     // Note: the GIL is used for synchronization of the singleton
-    d_instance_owner.reset();
+    s_instance_owner.reset();
     Py_RETURN_NONE;
 }
 
 Tracker*
 Tracker::getTracker()
 {
-    return d_instance;
+    return s_instance;
 }
 
 static struct
@@ -1132,26 +1150,31 @@ PyTraceFunction(
 }
 
 void
-forget_python_stack()
+Tracker::forgetPythonStack()
 {
     if (!Tracker::isActive()) {
         return;
     }
 
+    // Grab the Tracker lock, as this may need to write pushes/pops.
+    std::unique_lock<std::mutex> lock(*s_mutex);
     RecursionGuard guard;
     PythonStackTracker::get().clear();
 }
 
 void
-begin_tracking_greenlets()
+Tracker::beginTrackingGreenlets()
 {
     assert(PyGILState_Check());
     PythonStackTracker::s_greenlet_tracking_enabled = true;
 }
 
 void
-handle_greenlet_switch(PyObject* from, PyObject* to)
+Tracker::handleGreenletSwitch(PyObject* from, PyObject* to)
 {
+    // Grab the Tracker lock, as this may need to write pushes/pops.
+    std::unique_lock<std::mutex> lock(*s_mutex);
+    RecursionGuard guard;
     PythonStackTracker::get().handleGreenletSwitch(from, to);
 }
 
