@@ -9,6 +9,7 @@
 #include <fstream>
 #include <iterator>
 #include <memory>
+#include <optional>
 #include <string>
 #include <thread>
 #include <unordered_set>
@@ -93,36 +94,15 @@ PyTraceTrampoline(PyObject* obj, PyFrameObject* frame, int what, PyObject* arg);
 void
 install_trace_function();
 
-/**
- * Drop any references to frames on this thread's stack.
- *
- * This should be called when either the thread is dying or our profile
- * function is being uninstalled from it.
- */
-void
-forget_python_stack();
-
-/**
- * Sets a flag to enable integration with the `greenlet` module.
- */
-void
-begin_tracking_greenlets();
-
-/**
- * Handle a notification of control switching from one greenlet to another.
- */
-void
-handle_greenlet_switch(PyObject* from, PyObject* to);
-
 class NativeTrace
 {
   public:
     using ip_t = frame_id_t;
 
-    NativeTrace()
+    NativeTrace(std::vector<ip_t>& data)
+    : d_data(data)
     {
-        d_data.resize(MAX_SIZE);
-    };
+    }
 
     auto begin() const
     {
@@ -145,18 +125,17 @@ class NativeTrace
         size_t size;
         while (true) {
 #ifdef __linux__
-            size = unw_backtrace((void**)d_data.data(), MAX_SIZE);
+            size = unw_backtrace((void**)d_data.data(), d_data.size());
 #elif defined(__APPLE__)
-            size = ::backtrace((void**)d_data.data(), MAX_SIZE);
+            size = ::backtrace((void**)d_data.data(), d_data.size());
 #else
             return 0;
 #endif
-            if (size < MAX_SIZE) {
+            if (size < d_data.size()) {
                 break;
             }
 
-            MAX_SIZE = MAX_SIZE * 2;
-            d_data.resize(MAX_SIZE);
+            d_data.resize(d_data.size() * 2);
         }
         d_size = size > skip ? size - skip : 0;
         d_skip = skip;
@@ -189,13 +168,10 @@ class NativeTrace
 #endif
     }
 
-  private:
-    MEMRAY_FAST_TLS static thread_local size_t MAX_SIZE;
-
   private:
     size_t d_size = 0;
     size_t d_skip = 0;
-    std::vector<ip_t> d_data;
+    std::vector<ip_t>& d_data;
 };
 
 /**
@@ -232,15 +208,54 @@ class Tracker
     __attribute__((always_inline)) inline static void
     trackAllocation(void* ptr, size_t size, hooks::Allocator func)
     {
+        if (RecursionGuard::isActive || !Tracker::isActive()) {
+            return;
+        }
+        RecursionGuard guard;
+
+        std::optional<NativeTrace> trace{std::nullopt};
+        if (Tracker::areNativeTracesEnabled()) {
+            if (!prepareNativeTrace(trace)) {
+                return;
+            }
+            // Skip the internal frames so we don't need to filter them later.
+            trace.value().fill(1);
+        }
+
+        std::unique_lock<std::mutex> lock(*s_mutex);
         Tracker* tracker = getTracker();
         if (tracker) {
-            tracker->trackAllocationImpl(ptr, size, func);
+            tracker->trackAllocationImpl(ptr, size, func, trace);
         }
     }
 
+    static inline bool prepareNativeTrace(std::optional<NativeTrace>& trace)
+    {
+        auto t_trace_data_ptr = static_cast<std::vector<NativeTrace::ip_t>*>(
+                pthread_getspecific(s_native_unwind_vector_key));
+        if (!t_trace_data_ptr) {
+            t_trace_data_ptr = new std::vector<NativeTrace::ip_t>();
+            if (pthread_setspecific(s_native_unwind_vector_key, t_trace_data_ptr) != 0) {
+                Tracker::deactivate();
+                std::cerr << "memray: pthread_setspecific failed" << std::endl;
+                delete t_trace_data_ptr;
+                return false;
+            }
+            t_trace_data_ptr->resize(128);
+        }
+        trace.emplace(*t_trace_data_ptr);
+        return true;
+    }
+
     __attribute__((always_inline)) inline static void
     trackDeallocation(void* ptr, size_t size, hooks::Allocator func)
     {
+        if (RecursionGuard::isActive || !Tracker::isActive()) {
+            return;
+        }
+        RecursionGuard guard;
+
+        std::unique_lock<std::mutex> lock(*s_mutex);
         Tracker* tracker = getTracker();
         if (tracker) {
             tracker->trackDeallocationImpl(ptr, size, func);
@@ -249,22 +264,26 @@ class Tracker
 
     __attribute__((always_inline)) inline static void invalidate_module_cache()
     {
+        if (RecursionGuard::isActive || !Tracker::isActive()) {
+            return;
+        }
+        RecursionGuard guard;
+
+        std::unique_lock<std::mutex> lock(*s_mutex);
         Tracker* tracker = getTracker();
         if (tracker) {
             tracker->invalidate_module_cache_impl();
         }
     }
 
-    __attribute__((always_inline)) inline static void updateModuleCache()
-    {
-        Tracker* tracker = getTracker();
-        if (tracker) {
-            tracker->updateModuleCacheImpl();
-        }
-    }
-
     __attribute__((always_inline)) inline static void registerThreadName(const char* name)
     {
+        if (RecursionGuard::isActive || !Tracker::isActive()) {
+            return;
+        }
+        RecursionGuard guard;
+
+        std::unique_lock<std::mutex> lock(*s_mutex);
         Tracker* tracker = getTracker();
         if (tracker) {
             tracker->registerThreadNameImpl(name);
@@ -276,10 +295,28 @@ class Tracker
     bool popFrames(uint32_t count);
 
     // Interface to activate/deactivate the tracking
-    static const std::atomic<bool>& isActive();
+    static bool isActive();
     static void activate();
     static void deactivate();
 
+    /**
+     * Drop any references to frames on this thread's stack.
+     *
+     * This should be called when either the thread is dying or our profile
+     * function is being uninstalled from it.
+     */
+    static void forgetPythonStack();
+
+    /**
+     * Sets a flag to enable integration with the `greenlet` module.
+     */
+    static void beginTrackingGreenlets();
+
+    /**
+     * Handle a notification of control switching from one greenlet to another.
+     */
+    static void handleGreenletSwitch(PyObject* from, PyObject* to);
+
   private:
     class BackgroundThread
     {
@@ -307,17 +344,18 @@ class Tracker
     };
 
     // Data members
-    FrameCollection<RawFrame> d_frames;
-    static std::atomic<bool> d_active;
-    static std::unique_ptr<Tracker> d_instance_owner;
-    static std::atomic<Tracker*> d_instance;
+    static std::unique_ptr<std::mutex> s_mutex;
+    static pthread_key_t s_native_unwind_vector_key;
+    static std::unique_ptr<Tracker> s_instance_owner;
+    static std::atomic<Tracker*> s_instance;
 
+    FrameCollection<RawFrame> d_frames;
     std::shared_ptr<RecordWriter> d_writer;
     FrameTree d_native_trace_tree;
-    bool d_unwind_native_frames;
-    unsigned int d_memory_interval;
-    bool d_follow_fork;
-    bool d_trace_python_allocators;
+    const bool d_unwind_native_frames;
+    const unsigned int d_memory_interval;
+    const bool d_follow_fork;
+    const bool d_trace_python_allocators;
     linker::SymbolPatcher d_patcher;
     std::unique_ptr<BackgroundThread> d_background_thread;
 
@@ -325,7 +363,11 @@ class Tracker
     static size_t computeMainTidSkip();
     frame_id_t registerFrame(const RawFrame& frame);
 
-    void trackAllocationImpl(void* ptr, size_t size, hooks::Allocator func);
+    void trackAllocationImpl(
+            void* ptr,
+            size_t size,
+            hooks::Allocator func,
+            const std::optional<NativeTrace>& trace);
     void trackDeallocationImpl(void* ptr, size_t size, hooks::Allocator func);
     void invalidate_module_cache_impl();
     void updateModuleCacheImpl();
@@ -343,6 +385,8 @@ class Tracker
     static void prepareFork();
     static void parentFork();
     static void childFork();
+
+    static bool areNativeTracesEnabled();
 };
 
 }  // namespace memray::tracking_api
